///////////////////////////////////////////////////////////////////////////////
//                   SOFTWARE COPYRIGHT NOTICE AGREEMENT                     //
//       This software and its documentation are copyright (2012) by the     //
//   Broad Institute.  All rights are reserved.  This software is supplied   //
//   without any warranty or guaranteed support whatsoever. The Broad        //
//   Institute is not responsible for its use, misuse, or functionality.     //
///////////////////////////////////////////////////////////////////////////////
/*
 * \file Atomic.h
 * \author tsharpe
 * \date Jan 10, 2013
 *
 * \brief
 */
#ifndef ATOMIC_H_
#define ATOMIC_H_

// to be replaced by std::atomic momentarily, i hope
template <class T> // T is some kind of integer
class Atomic
{
public:
    Atomic( T initialVal = T() ) : mT(initialVal) {}
    // default move, copy, and destructor OK
    Atomic& operator=( T newVal ) { mT = newVal; }
    T operator++() { return changeVal(T(1))+1; }
    T operator++(int) { return changeVal(T(1)); }
    T operator--() { return changeVal(T(0)-1)-1; }
    T operator--(int) { return changeVal(T(0)-1); }
    T operator+=( T delta ) { return changeVal(delta)+delta; }
    T operator-=( T delta ) { return changeVal(-delta)-delta; }
    operator T() { return mT; }
private:
    T changeVal( T delta )
    { T oldVal, newVal;
      do
      {
          oldVal = mT;
          newVal = oldVal + delta;
      }
      while ( !__sync_bool_compare_and_swap(&mT,oldVal,newVal) );
      return oldVal;
    }
    T mT;
};

#endif /* ATOMIC_H_ */
